# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Collection

from sqlalchemy import any_, or_, select

from data_rentgen.db.models.dataset_symlink import DatasetSymlink, DatasetSymlinkType
from data_rentgen.db.repositories.base import Repository
from data_rentgen.dto.dataset_symlink import DatasetSymlinkDTO


class DatasetSymlinkRepository(Repository[DatasetSymlink]):
    async def create_or_update(self, dataset_symlink: DatasetSymlinkDTO) -> DatasetSymlink:
        result = await self._get(dataset_symlink)
        if not result:
            # try one more time, but with lock acquired.
            # if another worker already created the same row, just use it. if not - create with holding the lock.
            await self._lock(dataset_symlink.from_dataset.id, dataset_symlink.to_dataset.id)
            result = await self._get(dataset_symlink)

        if not result:
            return await self._create(dataset_symlink)
        return await self._update(result, dataset_symlink)

    async def list_by_dataset_ids(self, dataset_ids: Collection[int]) -> list[DatasetSymlink]:
        if not dataset_ids:
            return []

        query = select(DatasetSymlink).where(
            or_(
                DatasetSymlink.from_dataset_id == any_(list(dataset_ids)),  # type: ignore[arg-type]
                DatasetSymlink.to_dataset_id == any_(list(dataset_ids)),  # type: ignore[arg-type]
            ),
        )
        scalars = await self._session.scalars(query)
        return list(scalars.all())

    async def _get(self, dataset_symlink: DatasetSymlinkDTO) -> DatasetSymlink | None:
        query = select(DatasetSymlink).where(
            DatasetSymlink.from_dataset_id == dataset_symlink.from_dataset.id,
            DatasetSymlink.to_dataset_id == dataset_symlink.to_dataset.id,
        )
        return await self._session.scalar(query)

    async def _create(self, dataset_symlink: DatasetSymlinkDTO) -> DatasetSymlink:
        result = DatasetSymlink(
            from_dataset_id=dataset_symlink.from_dataset.id,
            to_dataset_id=dataset_symlink.to_dataset.id,
            type=DatasetSymlinkType(dataset_symlink.type),
        )
        self._session.add(result)
        await self._session.flush([result])
        return result

    async def _update(self, existing: DatasetSymlink, new: DatasetSymlinkDTO) -> DatasetSymlink:
        # almost of fields are immutable, so we can avoid UPDATE statements if row is unchanged
        existing.type = DatasetSymlinkType(new.type)
        await self._session.flush([existing])
        return existing



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from pydantic import Field

from data_rentgen.consumer.openlineage.dataset_facets.base import (
    OpenLineageInputDatasetFacet,
)


class OpenLineageInputStatisticsInputDatasetFacet(OpenLineageInputDatasetFacet):
    """Dataset facet describing Input statistics.
    See [InputStatisticsInputDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/InputStatisticsInputDatasetFacet.json).
    """

    rows: int | None = Field(default=None, alias="rowCount")
    bytes: int | None = Field(default=None, alias="size")
    files: int | None = Field(default=None, alias="fileCount")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from datetime import datetime
from enum import Enum, IntEnum
from uuid import UUID

from sqlalchemy import UUID as SQL_UUID
from sqlalchemy import DateTime, Integer, PrimaryKeyConstraint, SmallInteger, String
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy_utils import ChoiceType

from data_rentgen.db.models.base import Base
from data_rentgen.db.models.run import Run


class OperationStatus(IntEnum):
    UNKNOWN = -1
    """No data about status"""

    STARTED = 0
    """Received START event"""

    SUCCEEDED = 1
    """Finished successfully"""

    FAILED = 2
    """Internal failure"""

    KILLED = 3
    """Killed externally, e.g. by user request or in case of OOM"""


class OperationType(str, Enum):
    BATCH = "BATCH"
    STREAMING = "STREAMING"

    def __str__(self) -> str:
        return str(self.value)


class Operation(Base):
    __tablename__ = "operation"
    __table_args__ = (
        PrimaryKeyConstraint("created_at", "id"),
        {"postgresql_partition_by": "RANGE (created_at)"},
    )

    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        doc="Timestamp component of UUID, used for table partitioning",
    )
    id: Mapped[UUID] = mapped_column(SQL_UUID)

    run_id: Mapped[UUID] = mapped_column(
        SQL_UUID,
        index=True,
        nullable=False,
        doc="Run operation is a part of",
    )
    run: Mapped[Run] = relationship(
        Run,
        primaryjoin="Operation.run_id == Run.id",
        lazy="noload",
        foreign_keys=[run_id],
    )

    status: Mapped[OperationStatus] = mapped_column(
        ChoiceType(OperationStatus, impl=SmallInteger()),
        nullable=False,
        default=OperationStatus.UNKNOWN,
        doc="Operation status",
    )

    name: Mapped[str] = mapped_column(
        String,
        nullable=False,
        doc="Name of the operation, e.g. job name",
    )

    type: Mapped[OperationType] = mapped_column(
        ChoiceType(OperationType, impl=String(32)),
        nullable=False,
        default=OperationType.BATCH,
        doc="Type of the operation, e.g. BATCH, STREAMING",
    )

    position: Mapped[int | None] = mapped_column(
        Integer,
        nullable=True,
        doc="Sequentinal position of operation within the run, e.g. Spark jobId",
    )

    group: Mapped[str | None] = mapped_column(
        String,
        nullable=True,
        doc="Group of the operation, e.g. Spark jobGroup",
    )

    description: Mapped[str | None] = mapped_column(
        String,
        nullable=True,
        doc="Description of the operation, e.g. Spark jobDescription",
    )

    started_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True),
        nullable=True,
        doc="Start time of the operation",
    )
    ended_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True),
        nullable=True,
        doc="End time of the operation",
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from asgi_correlation_id import CorrelationIdMiddleware
from fastapi import FastAPI
from uuid6 import uuid8

from data_rentgen.server.settings.server import RequestIDSettings


def apply_request_id_middleware(app: FastAPI, settings: RequestIDSettings) -> FastAPI:
    """Add X-Request-ID middleware to the application."""
    if not settings.enabled:
        return app

    app.add_middleware(
        CorrelationIdMiddleware,
        generator=lambda: uuid8().hex,
        validator=None,
        **settings.model_dump(exclude={"enabled"}),
    )
    return app



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create schema

Revision ID: 81153f43e276
Revises: f9a114951102
Create Date: 2024-06-27 18:45:49.259710

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "81153f43e276"
down_revision = "f9a114951102"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "schema",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("digest", sa.UUID(), nullable=False),
        sa.Column("fields", sa.JSON(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__schema")),
    )
    op.create_index(op.f("ix__schema__digest"), "schema", ["digest"], unique=True)


def downgrade() -> None:
    op.drop_index(op.f("ix__schema__digest"), table_name="schema")
    op.drop_table("schema")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from typing import Literal

from faststream.security import SASLPlaintext
from pydantic import Field, SecretStr

from data_rentgen.consumer.settings.security.base import KafkaSecurityBaseSettings


class KafkaSecurityPlaintextSettings(KafkaSecurityBaseSettings):
    """Kafka PLAINTEXT auth settings.

    Examples
    --------

    .. code-block:: bash

        DATA_RENTGEN__KAFKA__SECURITY__TYPE=PLAINTEXT
        DATA_RENTGEN__KAFKA__SECURITY__USER=dummy
        DATA_RENTGEN__KAFKA__SECURITY__PASSWORD=changeme
    """

    type: Literal["PLAINTEXT"] = "PLAINTEXT"
    user: str = Field(description="Kafka security username")
    password: SecretStr = Field(description="Kafka security password")

    def to_security(self):
        return SASLPlaintext(
            self.user,
            self.password.get_secret_value(),
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from typing import Annotated

from fastapi import APIRouter, Depends, Request
from fastapi.responses import Response
from fastapi.security import OAuth2PasswordRequestForm

from data_rentgen.db.models.user import User
from data_rentgen.dependencies import Stub
from data_rentgen.server.errors.registration import get_error_responses
from data_rentgen.server.errors.schemas.invalid_request import InvalidRequestSchema
from data_rentgen.server.errors.schemas.logout import LogoutErrorSchema
from data_rentgen.server.errors.schemas.not_authorized import (
    NotAuthorizedRedirectSchema,
    NotAuthorizedSchema,
)
from data_rentgen.server.errors.schemas.not_implemented import NotImplementedErrorSchema
from data_rentgen.server.providers.auth import (
    AuthProvider,
    DummyAuthProvider,
    KeycloakAuthProvider,
)
from data_rentgen.server.schemas.v1.auth import AuthTokenSchema
from data_rentgen.server.services import get_user

router = APIRouter(
    prefix="/auth",
    tags=["Auth"],
    responses=get_error_responses(
        include={
            NotAuthorizedSchema,
            InvalidRequestSchema,
            NotAuthorizedRedirectSchema,
            LogoutErrorSchema,
            NotImplementedErrorSchema,
        },
    ),
)


@router.post("/token")
async def token(
    auth_provider: Annotated[DummyAuthProvider, Depends(Stub(AuthProvider))],
    form_data: Annotated[OAuth2PasswordRequestForm, Depends()],
) -> AuthTokenSchema:
    user_token = await auth_provider.get_token_password_grant(
        login=form_data.username,
        password=form_data.password,
    )
    return AuthTokenSchema.model_validate(user_token)


@router.get("/callback")
async def auth_callback(
    request: Request,
    code: str,
    auth_provider: Annotated[KeycloakAuthProvider, Depends(Stub(AuthProvider))],
):
    code_grant = await auth_provider.get_token_authorization_code_grant(
        code=code,
    )
    request.session["access_token"] = code_grant["access_token"]
    request.session["refresh_token"] = code_grant["refresh_token"]
    return Response(status_code=204)


@router.get("/logout")
async def logout(
    request: Request,
    current_user: Annotated[User, Depends(get_user())],
    auth_provider: Annotated[KeycloakAuthProvider, Depends(Stub(AuthProvider))],
):
    refresh_token = request.session.get("refresh_token", None)
    request.session.clear()
    await auth_provider.logout(user=current_user, refresh_token=refresh_token)
    return Response(status_code=204)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from data_rentgen.server.schemas.v1.address import AddressResponseV1
from data_rentgen.server.schemas.v1.dataset import (
    DatasetDetailedResponseV1,
    DatasetPaginateQueryV1,
    DatasetResponseV1,
)
from data_rentgen.server.schemas.v1.job import (
    JobDetailedResponseV1,
    JobPaginateQueryV1,
    JobResponseV1,
)
from data_rentgen.server.schemas.v1.lineage import (
    ColumnLineageInteractionTypeV1,
    DatasetLineageQueryV1,
    DirectLineageColumnRelationV1,
    IndirectLineageColumnRelationV1,
    JobLineageQueryV1,
    LineageDirectionV1,
    LineageEntityKindV1,
    LineageEntityV1,
    LineageInputRelationV1,
    LineageIORelationSchemaFieldV1,
    LineageIORelationSchemaV1,
    LineageNodesResponseV1,
    LineageOutputRelationV1,
    LineageParentRelationV1,
    LineageRelationsResponseV1,
    LineageResponseV1,
    LineageSourceColumnV1,
    LineageSymlinkRelationV1,
    OperationLineageQueryV1,
    RunLineageQueryV1,
)
from data_rentgen.server.schemas.v1.location import (
    LocationDetailedResponseV1,
    LocationPaginateQueryV1,
    LocationResponseV1,
    UpdateLocationRequestV1,
)
from data_rentgen.server.schemas.v1.operation import (
    OperationDetailedResponseV1,
    OperationIOStatisticsReponseV1,
    OperationQueryV1,
    OperationResponseV1,
    OperationStatisticsReponseV1,
)
from data_rentgen.server.schemas.v1.pagination import (
    PageMetaResponseV1,
    PageResponseV1,
    PaginateQueryV1,
)
from data_rentgen.server.schemas.v1.run import (
    RunDetailedResponseV1,
    RunIOStatisticsReponseV1,
    RunOperationStatisticsReponseV1,
    RunResponseV1,
    RunsQueryV1,
    RunStatisticsReponseV1,
)
from data_rentgen.server.schemas.v1.user import UserResponseV1

__all__ = [
    "AddressResponseV1",
    "ColumnLineageInteractionTypeV1",
    "DatasetDetailedResponseV1",
    "DatasetLineageQueryV1",
    "DatasetPaginateQueryV1",
    "DatasetResponseV1",
    "DirectLineageColumnRelationV1",
    "IndirectLineageColumnRelationV1",
    "JobDetailedResponseV1",
    "JobLineageQueryV1",
    "JobPaginateQueryV1",
    "JobResponseV1",
    "LineageDirectionV1",
    "LineageEntityKindV1",
    "LineageEntityV1",
    "LineageIORelationSchemaFieldV1",
    "LineageIORelationSchemaV1",
    "LineageInputRelationV1",
    "LineageNodesResponseV1",
    "LineageOutputRelationV1",
    "LineageParentRelationV1",
    "LineageRelationsResponseV1",
    "LineageResponseV1",
    "LineageSourceColumnV1",
    "LineageSymlinkRelationV1",
    "LocationDetailedResponseV1",
    "LocationPaginateQueryV1",
    "LocationResponseV1",
    "OperationDetailedResponseV1",
    "OperationIOStatisticsReponseV1",
    "OperationLineageQueryV1",
    "OperationQueryV1",
    "OperationResponseV1",
    "OperationStatisticsReponseV1",
    "PageMetaResponseV1",
    "PageResponseV1",
    "PaginateQueryV1",
    "RunDetailedResponseV1",
    "RunIOStatisticsReponseV1",
    "RunLineageQueryV1",
    "RunOperationStatisticsReponseV1",
    "RunResponseV1",
    "RunStatisticsReponseV1",
    "RunsQueryV1",
    "UpdateLocationRequestV1",
    "UserResponseV1",
]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from faststream import Depends, Logger
from sqlalchemy.ext.asyncio import AsyncSession

from data_rentgen.consumer.extractors import BatchExtractionResult, extract_batch
from data_rentgen.consumer.openlineage.run_event import OpenLineageRunEvent
from data_rentgen.dependencies import Stub
from data_rentgen.services.uow import UnitOfWork

__all__ = [
    "runs_events_subscriber",
]


def get_unit_of_work(session: AsyncSession = Depends(Stub(AsyncSession))) -> UnitOfWork:
    return UnitOfWork(session)


async def runs_events_subscriber(
    events: list[OpenLineageRunEvent],
    logger: Logger,
    unit_of_work: UnitOfWork = Depends(get_unit_of_work),
):
    logger.info("Got %d events", len(events))
    extracted = extract_batch(events)
    logger.info("Extracted: %r", extracted)

    logger.info("Saving to database")
    await save_to_db(extracted, unit_of_work, logger)
    logger.info("Saved successfully")


async def save_to_db(
    data: BatchExtractionResult,
    unit_of_work: UnitOfWork,
    logger: Logger,
) -> None:
    # To avoid deadlocks when parallel consumer instances insert/update the same row,
    # commit changes for each row instead of committing the whole batch. Yes, this cloud be slow.

    logger.debug("Creating locations")
    for location_dto in data.locations():
        async with unit_of_work:
            location = await unit_of_work.location.create_or_update(location_dto)
            location_dto.id = location.id

    logger.debug("Creating datasets")
    for dataset_dto in data.datasets():
        async with unit_of_work:
            dataset = await unit_of_work.dataset.create_or_update(dataset_dto)
            dataset_dto.id = dataset.id

    logger.debug("Creating symlinks")
    for dataset_symlink_dto in data.dataset_symlinks():
        async with unit_of_work:
            dataset_symlink = await unit_of_work.dataset_symlink.create_or_update(dataset_symlink_dto)
            dataset_symlink_dto.id = dataset_symlink.id

    logger.debug("Creating jobs")
    for job_dto in data.jobs():
        async with unit_of_work:
            job = await unit_of_work.job.create_or_update(job_dto)
            job_dto.id = job.id

    logger.debug("Creating users")
    for user_dto in data.users():
        async with unit_of_work:
            user = await unit_of_work.user.get_or_create(user_dto)
            user_dto.id = user.id

    logger.debug("Creating schemas")
    for schema_dto in data.schemas():
        async with unit_of_work:
            schema = await unit_of_work.schema.get_or_create(schema_dto)
            schema_dto.id = schema.id

    # Some events related to specific run are send to the same Kafka partition,
    # but at the same time we have parent_run which may be already inserted/updated by other worker
    # (Kafka key maybe different for run and it's parent).
    # In this case we cannot insert all the rows in one transaction, as it may lead to deadlocks.
    logger.debug("Creating runs")
    for run_dto in data.runs():
        async with unit_of_work:
            await unit_of_work.run.create_or_update(run_dto)

    # All events related to same operation are always send to the same Kafka partition,
    # so other workers never insert/update the same operation in parallel.
    # These rows can be inserted/updated in bulk, in one transaction.
    async with unit_of_work:
        logger.debug("Creating operations")
        await unit_of_work.operation.create_or_update_bulk(data.operations())

        logger.debug("Creating inputs")
        await unit_of_work.input.create_or_update_bulk(data.inputs())

        logger.debug("Creating outputs")
        await unit_of_work.output.create_or_update_bulk(data.outputs())

    # If something went wrong here, at least we will have inputs/outputs
    async with unit_of_work:
        column_lineage = data.column_lineage()
        logger.debug("Creating dataset column relations")
        await unit_of_work.dataset_column_relation.create_bulk_for_column_lineage(column_lineage)

        logger.debug("Creating column lineage")
        await unit_of_work.column_lineage.create_bulk(column_lineage)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from data_rentgen.dto.dataset import DatasetDTO
from data_rentgen.dto.operation import OperationDTO
from data_rentgen.dto.schema import SchemaDTO


@dataclass
class InputDTO:
    operation: OperationDTO
    dataset: DatasetDTO
    schema: SchemaDTO | None = None
    num_rows: int | None = None
    num_bytes: int | None = None
    num_files: int | None = None
    # id is generated using other ids combination
    id: UUID | None = None

    @property
    def unique_key(self) -> tuple:
        return (
            self.operation.unique_key,
            self.dataset.unique_key,
            (self.schema.unique_key if self.schema else None),
        )

    def merge(self, new: InputDTO) -> InputDTO:
        schema: SchemaDTO | None
        if self.schema and new.schema:  # noqa: SIM108
            schema = self.schema.merge(new.schema)
        else:
            schema = new.schema or self.schema

        return InputDTO(
            operation=self.operation.merge(new.operation),
            dataset=self.dataset.merge(new.dataset),
            schema=schema,
            num_rows=max(filter(None, [new.num_rows, self.num_rows]), default=None),
            num_bytes=max(filter(None, [new.num_bytes, self.num_bytes]), default=None),
            num_files=max(filter(None, [new.num_files, self.num_files]), default=None),
            id=new.id or self.id,
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from pydantic import BaseModel, Field, SecretStr


class KeycloakSettings(BaseModel):
    server_url: str = Field(..., description="Keycloak server URL")
    client_id: str = Field(..., description="Keycloak client ID")
    realm_name: str = Field(..., description="Keycloak realm name")
    client_secret: SecretStr = Field(..., description="Keycloak client secret")
    redirect_uri: str = Field(..., description="Redirect URI")
    verify_ssl: bool = Field(default=True, description="Verify SSL certificates")
    scope: str = Field("openid", description="Keycloak scope")


class KeycloakAuthProviderSettings(BaseModel):
    """Settings related to Keycloak interaction."

    Examples
    --------

    .. code-block:: bash

        DATA_RENTGEN__AUTH__PROVIDER=data_rentgen.server.providers.auth.keycloak_provider.KeycloakAuthProvider
        DATA_RENTGEN__AUTH__KEYCLOAK__SERVER_URL=http://keycloak:8080
        DATA_RENTGEN__AUTH__KEYCLOAK__REDIRECT_URI=http://localhost:8000/auth-callback
        DATA_RENTGEN__AUTH__KEYCLOAK__REALM_NAME=fastapi_realm
        DATA_RENTGEN__AUTH__KEYCLOAK__CLIENT_ID=fastapi_client
        DATA_RENTGEN__AUTH__KEYCLOAK__CLIENT_SECRET=6x6gn8uJdWSBmP8FqbNRSoGdvaoaFeez
        DATA_RENTGEN__AUTH__KEYCLOAK__SCOPE=email
        DATA_RENTGEN__AUTH__KEYCLOAK__VERIFY_SSL=False
    """

    keycloak: KeycloakSettings = Field(
        description="Keycloak settings",
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from pydantic import BaseModel


class TokenPayloadSchema(BaseModel):
    user_id: int


class AuthTokenSchema(BaseModel):
    access_token: str
    token_type: str
    expires_at: float



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from abc import ABC
from collections.abc import Sequence
from hashlib import sha1
from typing import Any, Generic, TypeVar

from sqlalchemy import (
    ColumnElement,
    CompoundSelect,
    ScalarResult,
    Select,
    SQLColumnExpression,
    func,
    select,
)
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.sql.base import ExecutableOption

from data_rentgen.db.models import Base
from data_rentgen.dto import PaginationDTO

Model = TypeVar("Model", bound=Base)


class Repository(ABC, Generic[Model]):
    def __init__(
        self,
        session: AsyncSession,
    ) -> None:
        self._session = session

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}()"

    @classmethod
    def model_type(cls) -> type[Model]:
        # Get `User` from `UserRepository(Repository[User])`
        return cls.__orig_bases__[0].__args__[0]  # type: ignore[attr-defined]

    async def _paginate_by_query(
        self,
        query: Select | CompoundSelect,
        order_by: Sequence[ColumnElement | SQLColumnExpression],
        page: int,
        page_size: int,
        options: Sequence[ExecutableOption] | None = None,
    ) -> PaginationDTO[Model]:
        model_type = self.model_type()
        items_query = select(model_type).from_statement(
            query.order_by(*order_by).limit(page_size).offset((page - 1) * page_size),
        )
        if options:
            items_query = items_query.options(*options)

        total_query = select(func.count()).select_from(query.subquery())
        items_result: ScalarResult[Model] = await self._session.scalars(items_query)

        total_count: int = await self._session.scalar(total_query)  # type: ignore[assignment]
        return PaginationDTO[model_type](  # type: ignore[valid-type]
            items=list(items_result.all()),
            total_count=total_count,
            page=page,
            page_size=page_size,
        )

    async def _lock(
        self,
        *keys: Any,
    ) -> None:
        """
        Take a lock on a specific table and set of keys, to avoid inserting the same row multiple times.
        Based on [pg_advisory_xact_lock](https://www.postgresql.org/docs/current/functions-admin.html).
        Lock is held until the transaction is committed or rolled back.
        """
        model_type = self.model_type()
        data = ".".join(map(str, [model_type.__table__, *keys]))
        digest = sha1(data.encode("utf-8"), usedforsecurity=False).digest()
        # sha1 returns 160bit hash, we need only first 64 bits
        lock_key = int.from_bytes(digest[:8], byteorder="big", signed=True)
        statement = select(func.pg_advisory_xact_lock(lock_key))
        await self._session.execute(statement)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Collection
from datetime import datetime
from uuid import UUID

from sqlalchemy import (
    ColumnElement,
    CompoundSelect,
    Select,
    SQLColumnExpression,
    any_,
    desc,
    func,
    select,
    union,
)
from sqlalchemy.orm import selectinload

from data_rentgen.db.models import Job, Run, RunStartReason, RunStatus
from data_rentgen.db.repositories.base import Repository
from data_rentgen.db.utils.search import make_tsquery, ts_match, ts_rank
from data_rentgen.db.utils.uuid import extract_timestamp_from_uuid
from data_rentgen.dto import PaginationDTO, RunDTO


class RunRepository(Repository[Run]):
    async def create_or_update(self, run: RunDTO) -> Run:
        # avoid calculating created_at twice
        created_at = extract_timestamp_from_uuid(run.id)
        result = await self._get(created_at, run.id)
        if not result:
            # try one more time, but with lock acquired.
            # if another worker already created the same row, just use it. if not - create with holding the lock.
            await self._lock(run.id)
            result = await self._get(created_at, run.id)

        if not result:
            return await self._create(created_at, run)
        return await self._update(result, run)

    async def paginate(
        self,
        page: int,
        page_size: int,
        since: datetime | None,
        until: datetime | None,
        run_ids: Collection[UUID],
        job_id: int | None,
        parent_run_id: UUID | None,
        search_query: str | None,
    ) -> PaginationDTO[Run]:
        # do not use `tuple_(Run.created_at, Run.id).in_(...),
        # as this is too complex filter for Postgres to make an optimal query plan
        where = []
        if run_ids:
            min_run_created_at = extract_timestamp_from_uuid(min(run_ids))
            max_run_created_at = extract_timestamp_from_uuid(max(run_ids))
            min_created_at = max(since, min_run_created_at) if since else min_run_created_at
            max_created_at = min(until, max_run_created_at) if until else max_run_created_at
            where = [
                Run.created_at >= min_created_at,
                Run.created_at <= max_created_at,
                Run.id == any_(list(run_ids)),  # type: ignore[arg-type]
            ]
        else:
            if since:
                where.append(Run.created_at >= since)
            if until:
                where.append(Run.created_at <= until)

        if run_ids:
            where.append(Run.id == any_(list(run_ids)))  # type: ignore[arg-type]
        if job_id:
            where.append(Run.job_id == job_id)
        if parent_run_id:
            where.append(Run.parent_run_id == parent_run_id)

        query: Select | CompoundSelect
        order_by: list[ColumnElement | SQLColumnExpression]
        if search_query:
            tsquery = make_tsquery(search_query)

            run_stmt = select(Run, ts_rank(Run.search_vector, tsquery).label("search_rank")).where(
                ts_match(Run.search_vector, tsquery),
                *where,
            )
            job_stmt = (
                select(Run, ts_rank(Job.search_vector, tsquery).label("search_rank"))
                .join(Job, Job.id == Run.job_id)
                .where(ts_match(Job.search_vector, tsquery), *where)
            )

            union_cte = union(run_stmt, job_stmt).cte()

            run_columns = [column for column in union_cte.columns if column.name != "search_rank"]

            query = select(
                *run_columns,
                func.max(union_cte.c.search_rank).label("search_rank"),
            ).group_by(*run_columns)
            # place the most recent runs on top
            order_by = [desc("search_rank"), desc("created_at"), desc("id")]
        else:
            query = select(Run).where(*where)
            order_by = [Run.created_at.desc(), Run.id.desc()]

        options = [selectinload(Run.started_by_user)]
        return await self._paginate_by_query(
            query=query,
            order_by=order_by,
            options=options,
            page=page,
            page_size=page_size,
        )

    async def list_by_ids(self, run_ids: Collection[UUID]) -> list[Run]:
        if not run_ids:
            return []
        # do not use `tuple_(Run.created_at, Run.id).in_(...),
        # as this is too complex filter for Postgres to make an optimal query plan
        min_created_at = extract_timestamp_from_uuid(min(run_ids))
        max_created_at = extract_timestamp_from_uuid(max(run_ids))
        query = (
            select(Run)
            .where(
                Run.created_at >= min_created_at,
                Run.created_at <= max_created_at,
                Run.id == any_(list(run_ids)),  # type: ignore[arg-type]
            )
            .options(selectinload(Run.started_by_user))
        )
        result = await self._session.scalars(query)
        return list(result.all())

    async def list_by_job_ids(self, job_ids: Collection[int], since: datetime, until: datetime | None) -> list[Run]:
        if not job_ids:
            return []
        query = (
            select(Run)
            .where(
                Run.created_at >= since,
                Run.job_id == any_(list(job_ids)),  # type: ignore[arg-type]
            )
            .options(selectinload(Run.started_by_user))
        )
        if until:
            query = query.where(Run.created_at <= until)
        result = await self._session.scalars(query)
        return list(result.all())

    async def _get(self, created_at: datetime, run_id: UUID) -> Run | None:
        query = select(Run).where(Run.id == run_id, Run.created_at == created_at)
        return await self._session.scalar(query)

    async def _create(
        self,
        created_at: datetime,
        run: RunDTO,
    ) -> Run:
        result = Run(
            created_at=created_at,
            id=run.id,
            job_id=run.job.id,
            status=RunStatus(run.status),
            parent_run_id=run.parent_run.id if run.parent_run else None,
            started_at=run.started_at,
            started_by_user_id=run.user.id if run.user else None,
            start_reason=RunStartReason(run.start_reason) if run.start_reason else None,
            ended_at=run.ended_at,
            external_id=run.external_id,
            attempt=run.attempt,
            persistent_log_url=run.persistent_log_url,
            running_log_url=run.running_log_url,
        )
        self._session.add(result)
        await self._session.flush([result])
        return result

    async def _update(
        self,
        existing: Run,
        new: RunDTO,
    ) -> Run:
        # for parent_run most of fields are None, so we can avoid UPDATE statements if row is unchanged
        optional_fields = {
            # in some cases start event may be send from "unknown" job,
            # but sequential RUNNING and STOPPED events are send with proper job name. Rebound run to latest job.
            "job_id": new.job.id or existing.job_id,
            # Merge new information with existing one
            "status": max(RunStatus(new.status), existing.status),
            "parent_run_id": new.parent_run.id if new.parent_run else None,
            "started_at": new.started_at,
            "started_by_user_id": new.user.id if new.user else None,
            "start_reason": RunStartReason(new.start_reason) if new.start_reason else None,
            "ended_at": new.ended_at,
            "external_id": new.external_id,
            "attempt": new.attempt,
            "persistent_log_url": new.persistent_log_url,
            "running_log_url": new.running_log_url,
        }
        for column, value in optional_fields.items():
            if value is not None:
                setattr(existing, column, value)

        await self._session.flush([existing])
        return existing



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Collection
from dataclasses import dataclass
from datetime import datetime
from typing import Annotated
from uuid import UUID

from fastapi import Depends
from sqlalchemy import Row

from data_rentgen.db.models.operation import Operation
from data_rentgen.dto.pagination import PaginationDTO
from data_rentgen.services.uow import UnitOfWork


@dataclass
class OperationServiceIOStatistics:
    total_datasets: int = 0
    total_bytes: int = 0
    total_rows: int = 0
    total_files: int = 0

    @classmethod
    def from_row(cls, row: Row | None):
        if not row:
            return cls()

        return cls(
            total_datasets=row.total_datasets,
            total_bytes=row.total_bytes or 0,
            total_rows=row.total_rows or 0,
            total_files=row.total_files or 0,
        )


@dataclass
class OperationServiceStatistics:
    inputs: OperationServiceIOStatistics
    outputs: OperationServiceIOStatistics


@dataclass
class OperationServicePageItem:
    id: UUID
    data: Operation
    statistics: OperationServiceStatistics


class OperationServicePaginatedResult(PaginationDTO[OperationServicePageItem]):
    pass


class OperationService:
    def __init__(self, uow: Annotated[UnitOfWork, Depends()]):
        self._uow = uow

    async def paginate(
        self,
        page: int,
        page_size: int,
        since: datetime | None,
        until: datetime | None,
        operation_ids: Collection[UUID],
        run_id: UUID | None,
    ) -> OperationServicePaginatedResult:
        pagination = await self._uow.operation.paginate(
            page=page,
            page_size=page_size,
            since=since,
            until=until,
            operation_ids=operation_ids,
            run_id=run_id,
        )
        operation_ids = [item.id for item in pagination.items]
        input_stats = await self._uow.input.get_stats_by_operation_ids(operation_ids)
        output_stats = await self._uow.output.get_stats_by_operation_ids(operation_ids)

        return OperationServicePaginatedResult(
            page=pagination.page,
            page_size=pagination.page_size,
            total_count=pagination.total_count,
            items=[
                OperationServicePageItem(
                    id=operation.id,
                    data=operation,
                    statistics=OperationServiceStatistics(
                        inputs=OperationServiceIOStatistics.from_row(input_stats.get(operation.id)),
                        outputs=OperationServiceIOStatistics.from_row(output_stats.get(operation.id)),
                    ),
                )
                for operation in pagination.items
            ],
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import json
from dataclasses import dataclass, field
from functools import cached_property


@dataclass
class SchemaDTO:
    fields: list[dict]
    id: int | None = field(default=None, compare=False)

    @cached_property
    def unique_key(self) -> tuple:
        # expensive operation, calculate it only once
        return (json.dumps(self.fields, sort_keys=True),)

    def merge(self, new: SchemaDTO) -> SchemaDTO:
        if new.id is None:
            return self

        return SchemaDTO(
            fields=self.fields,
            id=new.id or self.id,
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from typing import Annotated, Union

from pydantic import Discriminator

from data_rentgen.consumer.settings.security.anonymous import KafkaSecurityAnonymousSettings
from data_rentgen.consumer.settings.security.gssapi import KafkaSecurityGSSAPISettings
from data_rentgen.consumer.settings.security.plain import KafkaSecurityPlaintextSettings
from data_rentgen.consumer.settings.security.scram import KafkaSecurityScram256Settings, KafkaSecurityScram512Settings

KafkaSecuritySettings = Annotated[
    Union[
        KafkaSecurityAnonymousSettings,
        KafkaSecurityScram256Settings,
        KafkaSecurityScram512Settings,
        KafkaSecurityPlaintextSettings,
        KafkaSecurityGSSAPISettings,
    ],
    Discriminator("type"),
]

__all__ = [
    "KafkaSecurityAnonymousSettings",
    "KafkaSecurityGSSAPISettings",
    "KafkaSecurityPlaintextSettings",
    "KafkaSecurityScram256Settings",
    "KafkaSecurityScram512Settings",
    "KafkaSecuritySettings",
]



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import asyncio
import logging
import sys
from argparse import ArgumentParser
from enum import Enum

from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession

from data_rentgen.db.factory import create_session_factory
from data_rentgen.db.settings import DatabaseSettings
from data_rentgen.logging.settings import LoggingSettings
from data_rentgen.logging.setup_logging import setup_logging

logger = logging.getLogger(__name__)


class Depth(str, Enum):
    DAY = "day"
    WEEK = "week"
    MONTH = "month"

    def __str__(self) -> str:
        return self.value


view_sufix_map = {
    "day": "_daily_stats",
    "week": "_weekly_stats",
    "month": "_monthly_stats",
}


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        usage="python3 -m data_rentgen.db.scripts.create_analytics_view --depths day",
        description="Create matherialized views based on input and output table with given depths",
    )
    parser.add_argument(
        "--depths",
        type=Depth,
        choices=[item.value for item in Depth],
        nargs="+",
        help="Depth of matherialized view data (created_at filter). You can provide list of args",
    )
    return parser


async def refresh_view(depth: Depth, session: AsyncSession):
    for base_table in ("output", "input"):
        view_name = base_table + view_sufix_map[depth]
        logger.info("Refresh view: %s", view_name)
        statement = f"REFRESH MATERIALIZED VIEW {view_name}"
        await session.execute(text(statement))


async def main(args: list[str]) -> None:
    setup_logging(LoggingSettings())

    parser = get_parser()
    params = parser.parse_args(args)
    depths = params.depths
    if not depths:
        logger.info("Create views for all depths")
        depths = Depth
    else:
        depths = sorted(set(depths))

    db_settings = DatabaseSettings()  # type: ignore[call-arg]
    session_factory = create_session_factory(db_settings)
    async with session_factory() as session:
        for depth in depths:
            await refresh_view(depth, session)
        await session.commit()


if __name__ == "__main__":
    asyncio.run(main(sys.argv[1:]))



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING

from fastapi import FastAPI
from sqlalchemy.ext.asyncio import AsyncSession

import data_rentgen
from data_rentgen.db.factory import session_generator
from data_rentgen.logging.setup_logging import setup_logging
from data_rentgen.server.api.handlers import apply_exception_handlers
from data_rentgen.server.api.router import api_router
from data_rentgen.server.middlewares import apply_middlewares
from data_rentgen.server.settings import ServerApplicationSettings

if TYPE_CHECKING:
    from data_rentgen.server.providers.auth import AuthProvider


def application_factory(settings: ServerApplicationSettings) -> FastAPI:
    application = FastAPI(
        title="Data.Rentgen",
        description="Data.Rentgen is a nextgen DataLineage service",
        version=data_rentgen.__version__,
        debug=settings.server.debug,
        # will be set up by middlewares
        openapi_url=None,
        docs_url=None,
        redoc_url=None,
    )

    application.state.settings = settings
    application.include_router(api_router)

    apply_exception_handlers(application)
    auth_class: type[AuthProvider] = settings.auth.provider  # type: ignore[assignment]
    auth_class.setup(application)
    apply_middlewares(application, settings.server)

    application.dependency_overrides.update(
        {
            ServerApplicationSettings: lambda: settings,
            AsyncSession: session_generator(settings.database),  # type: ignore[dict-item]
        },
    )
    return application


def get_application():
    settings = ServerApplicationSettings()
    setup_logging(settings.logging)
    return application_factory(settings=settings)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create job

Revision ID: 026de1556610
Revises: c2c583123133
Create Date: 2024-06-27 19:11:39.585904

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "026de1556610"
down_revision = "c2c583123133"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "job",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("location_id", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("type", sa.String(length=32), nullable=False),
        sa.Column(
            "search_vector",
            postgresql.TSVECTOR(),
            sa.Computed(
                "to_tsvector('simple'::regconfig,  name || ' ' || (translate(name, '/.', '  ')))",
                persisted=True,
            ),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["location_id"],
            ["location.id"],
            name=op.f("fk__job__location_id__location"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__job")),
        sa.UniqueConstraint("location_id", "name", name=op.f("uq__job__location_id_name")),
    )
    op.create_index(op.f("ix__job__location_id"), "job", ["location_id"], unique=False)
    op.create_index(op.f("ix__job__name"), "job", ["name"], unique=False)
    op.create_index(op.f("ix__job__type"), "job", ["type"], unique=False)
    op.create_index(op.f("ix__job__search_vector"), "job", ["search_vector"], unique=False, postgresql_using="gin")


def downgrade() -> None:
    op.drop_index(op.f("ix__job__type"), table_name="job")
    op.drop_index(op.f("ix__job__name"), table_name="job")
    op.drop_index(op.f("ix__job__location_id"), table_name="job")
    op.drop_index(op.f("ix__job_search__vector"), table_name="job", postgresql_using="gin")
    op.drop_table("job")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.exceptions.auth import ActionNotAllowedError, AuthorizationError, LogoutError
from data_rentgen.exceptions.base import ApplicationError
from data_rentgen.exceptions.entity import EntityNotFoundError
from data_rentgen.exceptions.redirect import RedirectError

__all__ = [
    "ActionNotAllowedError",
    "ApplicationError",
    "AuthorizationError",
    "EntityNotFoundError",
    "LogoutError",
    "RedirectError",
]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

import textwrap
from pathlib import Path

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Literal


class LoggingSettings(BaseSettings):
    """Data.Rentgen backend logging Settings.

    Examples
    --------

    Using ``json`` preset:

    .. code-block:: bash

        DATA_RENTGEN__LOGGING__SETUP=True
        DATA_RENTGEN__LOGGING__PRESET=json

    Passing custom logging config file:

    .. code-block:: bash

        DATA_RENTGEN__LOGGING__SETUP=True
        DATA_RENTGEN__LOGGING__CUSTOM_CONFIG_PATH=/some/logging.yml

    Setup logging in some other way, e.g. using `uvicorn args <https://www.uvicorn.org/settings/#logging>`_:

    .. code-block:: bash

        $ export DATA_RENTGEN__LOGGING__SETUP=False
        $ python -m data_rentgen.server --log-level debug
    """

    setup: bool = Field(
        default=True,
        description="If ``True``, setup logging during application start",
    )
    preset: Literal["json", "plain", "colored"] = Field(
        default="plain",
        description=textwrap.dedent(
            """
            Name of logging preset to use.

            There are few logging presets bundled to ``data-rentgen[server]`` package:

            .. dropdown:: ``plain`` preset

                This preset is recommended to use in environment which do not support colored output,
                e.g. CI jobs

                .. literalinclude:: ../../../../data_rentgen/logging/presets/plain.yml

            .. dropdown:: ``colored`` preset

                This preset is recommended to use in development environment,
                as it simplifies debugging. Each log record is output with color specific for a log level

                .. literalinclude:: ../../../../data_rentgen/logging/presets/colored.yml

            .. dropdown:: ``json`` preset

                This preset is recommended to use in production environment,
                as it allows to avoid writing complex log parsing configs. Each log record is output as JSON line

                .. literalinclude:: ../../../../data_rentgen/logging/presets/json.yml
            """,
        ),
    )

    custom_config_path: Path | None = Field(  # noqa: FA102
        default=None,
        description=textwrap.dedent(
            """
            Path to custom logging configuration file. If set, overrides :obj:`~preset` value.

            File content should be in YAML format and conform
            `logging.dictConfig <https://docs.python.org/3/library/logging.config.html#logging-config-dictschema>`_.
            """,
        ),
    )

    model_config = SettingsConfigDict(env_prefix="DATA_RENTGEN__LOGGING__", env_nested_delimiter="__", extra="forbid")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from typing import Annotated

from fastapi import Depends
from sqlalchemy.ext.asyncio import AsyncSession

from data_rentgen.db.repositories.column_lineage import ColumnLineageRepository
from data_rentgen.db.repositories.dataset import DatasetRepository
from data_rentgen.db.repositories.dataset_column_relation import (
    DatasetColumnRelationRepository,
)
from data_rentgen.db.repositories.dataset_symlink import DatasetSymlinkRepository
from data_rentgen.db.repositories.input import InputRepository
from data_rentgen.db.repositories.job import JobRepository
from data_rentgen.db.repositories.location import LocationRepository
from data_rentgen.db.repositories.operation import OperationRepository
from data_rentgen.db.repositories.output import OutputRepository
from data_rentgen.db.repositories.run import RunRepository
from data_rentgen.db.repositories.schema import SchemaRepository
from data_rentgen.db.repositories.user import UserRepository
from data_rentgen.dependencies import Stub


class UnitOfWork:
    def __init__(
        self,
        session: Annotated[AsyncSession, Depends(Stub(AsyncSession))],
    ):
        self._session = session
        self.location = LocationRepository(session)
        self.job = JobRepository(session)
        self.run = RunRepository(session)
        self.operation = OperationRepository(session)
        self.dataset = DatasetRepository(session)
        self.dataset_symlink = DatasetSymlinkRepository(session)
        self.schema = SchemaRepository(session)
        self.input = InputRepository(session)
        self.output = OutputRepository(session)
        self.dataset_column_relation = DatasetColumnRelationRepository(session)
        self.column_lineage = ColumnLineageRepository(session)
        self.user = UserRepository(session)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            await self._session.rollback()
        else:
            await self._session.commit()



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from datetime import datetime
from uuid import UUID

from sqlalchemy import UUID as SQL_UUID
from sqlalchemy import (
    BigInteger,
    DateTime,
    PrimaryKeyConstraint,
)
from sqlalchemy.orm import Mapped, mapped_column, relationship

from data_rentgen.db.models.base import Base
from data_rentgen.db.models.dataset import Dataset
from data_rentgen.db.models.dataset_column_relation import DatasetColumnRelation
from data_rentgen.db.models.job import Job
from data_rentgen.db.models.operation import Operation
from data_rentgen.db.models.run import Run


# no foreign keys to avoid scanning all the partitions
class ColumnLineage(Base):
    __tablename__ = "column_lineage"
    __table_args__ = (
        PrimaryKeyConstraint("created_at", "id"),
        {"postgresql_partition_by": "RANGE (created_at)"},
    )

    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        doc="Timestamp component of UUID, used for table partitioning",
    )
    id: Mapped[UUID] = mapped_column(SQL_UUID)

    operation_id: Mapped[UUID] = mapped_column(
        SQL_UUID,
        index=True,
        nullable=False,
        doc="Operation caused this column lineage",
    )
    operation: Mapped[Operation] = relationship(
        Operation,
        primaryjoin="ColumnLineage.operation_id == Operation.id",
        lazy="noload",
        foreign_keys=[operation_id],
    )

    run_id: Mapped[UUID] = mapped_column(
        SQL_UUID,
        index=True,
        nullable=False,
        doc="Run the operation is bound to",
    )
    run: Mapped[Run] = relationship(
        Run,
        primaryjoin="ColumnLineage.run_id == Run.id",
        lazy="noload",
        foreign_keys=[run_id],
    )

    job_id: Mapped[int] = mapped_column(
        BigInteger,
        index=True,
        nullable=False,
        doc="Parent job of run",
    )
    job: Mapped[Job] = relationship(
        Job,
        primaryjoin="ColumnLineage.job_id == Job.id",
        lazy="noload",
        foreign_keys=[job_id],
    )

    source_dataset_id: Mapped[int] = mapped_column(
        BigInteger,
        index=True,
        nullable=False,
        doc="Dataset the data is originated from",
    )
    source_dataset: Mapped[Dataset] = relationship(
        Dataset,
        primaryjoin="ColumnLineage.source_dataset_id == Dataset.id",
        lazy="noload",
        foreign_keys=[source_dataset_id],
    )

    target_dataset_id: Mapped[int] = mapped_column(
        BigInteger,
        index=True,
        nullable=False,
        doc="Dataset the data is saved to",
    )
    target_dataset: Mapped[Dataset] = relationship(
        Dataset,
        primaryjoin="ColumnLineage.target_dataset_id == Dataset.id",
        lazy="noload",
        foreign_keys=[target_dataset_id],
    )

    fingerprint: Mapped[UUID] = mapped_column(
        SQL_UUID,
        index=False,
        nullable=False,
        doc="Dataset column relation fingerprint",
    )
    dataset_column_relations: Mapped[list[DatasetColumnRelation]] = relationship(
        DatasetColumnRelation,
        uselist=True,
        primaryjoin="ColumnLineage.fingerprint == DatasetColumnRelation.fingerprint",
        lazy="noload",
        foreign_keys=[fingerprint],
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from datetime import datetime
from enum import Enum, IntEnum
from typing import Literal
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, ValidationInfo, field_serializer, field_validator

from data_rentgen.server.schemas.v1.dataset import DatasetResponseV1
from data_rentgen.server.schemas.v1.job import JobResponseV1
from data_rentgen.server.schemas.v1.operation import OperationResponseV1
from data_rentgen.server.schemas.v1.run import RunResponseV1
from data_rentgen.utils import UUIDv6Plus


class LineageEntityKindV1(str, Enum):
    JOB = "JOB"
    RUN = "RUN"
    OPERATION = "OPERATION"
    DATASET = "DATASET"

    def __str__(self) -> str:
        return self.value


class LineageDirectionV1(str, Enum):
    DOWNSTREAM = "DOWNSTREAM"
    UPSTREAM = "UPSTREAM"
    BOTH = "BOTH"

    def __str__(self) -> str:
        return self.value


class LineageEntityV1(BaseModel):
    kind: LineageEntityKindV1 = Field(description="Type of Lineage entity")
    id: str | UUID = Field(description="Id of Lineage entity")

    model_config = ConfigDict(from_attributes=True)


class BaseLineageQueryV1(BaseModel):
    since: datetime = Field(
        description="",
        examples=["2008-09-15T15:53:00+05:00"],
    )
    until: datetime | None = Field(
        default=None,
        description="",
        examples=["2008-09-15T15:53:00+05:00"],
    )
    direction: LineageDirectionV1 = Field(
        default=LineageDirectionV1.BOTH,
        description="Direction of the lineage",
        examples=["DOWNSTREAM", "UPSTREAM", "BOTH"],
    )
    depth: int = Field(
        default=1,
        ge=1,
        le=10,
        description="Depth of the lineage",
        examples=[1, 3],
    )
    include_column_lineage: bool = Field(default=False, description="Include or not column lineage into response")

    model_config = ConfigDict(extra="forbid")

    @field_validator("until", mode="after")
    @classmethod
    def _check_until(cls, value: datetime | None, info: ValidationInfo) -> datetime | None:
        since = info.data.get("since")
        if since and value and since >= value:
            msg = "'since' should be less than 'until'"
            raise ValueError(msg)
        return value


class DatasetLineageQueryV1(BaseLineageQueryV1):
    start_node_id: int = Field(description="Dataset id", examples=[42])
    granularity: Literal["OPERATION", "RUN", "JOB"] = Field(
        default="RUN",
        description="Granularity of the dataset lineage",
        examples=["OPERATION", "RUN", "JOB"],
    )

    model_config = ConfigDict(extra="forbid")


class JobLineageQueryV1(BaseLineageQueryV1):
    start_node_id: int = Field(description="Job id", examples=[42])
    granularity: Literal["JOB", "RUN"] = Field(
        default="JOB",
        description="Granularity of the job lineage",
        examples=["JOB", "RUN"],
    )


class OperationLineageQueryV1(BaseLineageQueryV1):
    start_node_id: UUIDv6Plus = Field(description="Operation id", examples=["00000000-0000-0000-0000-000000000000"])


class RunLineageQueryV1(BaseLineageQueryV1):
    start_node_id: UUIDv6Plus = Field(description="Run id", examples=["00000000-0000-0000-0000-000000000000"])
    granularity: Literal["OPERATION", "RUN"] = Field(
        default="RUN",
        description="Granularity of the run lineage",
        examples=["OPERATION", "RUN"],
    )


class LineageParentRelationV1(BaseModel):
    from_: LineageEntityV1 = Field(description="Start point of relation", serialization_alias="from")
    to: LineageEntityV1 = Field(description="End point of relation")


class LineageIORelationSchemaFieldV1(BaseModel):
    name: str
    type: str | None = Field(default=None)
    description: str | None = Field(default=None)
    fields: list[LineageIORelationSchemaFieldV1] = Field(description="Nested fields", default_factory=list)

    model_config = ConfigDict(from_attributes=True)


class LineageIORelationSchemaV1(BaseModel):
    id: str = Field(description="Schema id", coerce_numbers_to_str=True)
    fields: list[LineageIORelationSchemaFieldV1] = Field(description="Schema fields")
    relevance_type: Literal["EXACT_MATCH", "LATEST_KNOWN"] | None = Field(
        description="Relevance of schema",
        default="LATEST_KNOWN",
    )
    model_config = ConfigDict(from_attributes=True)


class LineageInputRelationV1(BaseModel):
    from_: LineageEntityV1 = Field(description="Start point of relation", serialization_alias="from")
    to: LineageEntityV1 = Field(description="End point of relation")
    last_interaction_at: datetime = Field(description="Last interaction at", examples=["2008-09-15T15:53:00+05:00"])
    num_bytes: int | None = Field(description="Number of bytes", examples=[42], default=None)
    num_rows: int | None = Field(description="Number of rows", examples=[42], default=None)
    num_files: int | None = Field(description="Number of files", examples=[42], default=None)
    i_schema: LineageIORelationSchemaV1 | None = Field(
        description="Schema",
        default=None,
        # pydantic models have reserved "schema" attribute, using alias
        serialization_alias="schema",
    )


class LineageOutputRelationV1(BaseModel):
    from_: LineageEntityV1 = Field(description="Start point of relation", serialization_alias="from")
    to: LineageEntityV1 = Field(description="End point of relation")
    type: str | None = Field(description="Type of relation", examples=["CREATE", "APPEND"], default=None)
    last_interaction_at: datetime = Field(description="Last interaction at", examples=["2008-09-15T15:53:00+05:00"])
    num_bytes: int | None = Field(description="Number of bytes", examples=[42], default=None)
    num_rows: int | None = Field(description="Number of rows", examples=[42], default=None)
    num_files: int | None = Field(description="Number of files", examples=[42], default=None)
    o_schema: LineageIORelationSchemaV1 | None = Field(
        description="Schema",
        default=None,
        # pydantic models have reserved "schema" attribute, using alias
        serialization_alias="schema",
    )


class LineageSymlinkRelationV1(BaseModel):
    from_: LineageEntityV1 = Field(description="Start point of relation", serialization_alias="from")
    to: LineageEntityV1 = Field(description="End point of relation")
    type: str = Field(description="Type of relation between datasets", examples=["METASTORE", "WAREHOUSE"])


class ColumnLineageInteractionTypeV1(IntEnum):
    UNKNOWN = 1

    # Direct
    IDENTITY = 2
    TRANSFORMATION = 4
    TRANSFORMATION_MASKING = 8
    AGGREGATION = 16
    AGGREGATION_MASKING = 32

    # Indirect
    FILTER = 64
    JOIN = 128
    GROUP_BY = 256
    SORT = 512
    WINDOW = 1024
    CONDITIONAL = 2048

    def __str__(self) -> str:
        return self.name


class LineageSourceColumnV1(BaseModel):
    field: str = Field(description="Source column", examples=["my_col_1", "my_col_2"])
    types: list[ColumnLineageInteractionTypeV1] = Field(
        description="Type of relation between source and target column",
        examples=[
            [
                ColumnLineageInteractionTypeV1.UNKNOWN,
            ],
            [
                ColumnLineageInteractionTypeV1.TRANSFORMATION,
                ColumnLineageInteractionTypeV1.AGGREGATION,
            ],
            [
                ColumnLineageInteractionTypeV1.JOIN,
            ],
        ],
    )
    last_used_at: datetime = Field(
        description="Last time when transformation was used at",
        examples=["2008-09-15T15:53:00+05:00"],
    )

    @field_serializer("types")
    def serialize_types(self, types: list[ColumnLineageInteractionTypeV1], _info):
        return [type_.name for type_ in types]


class DirectLineageColumnRelationV1(BaseModel):
    from_: LineageEntityV1 = Field(description="Dataset with source columns", serialization_alias="from")
    to: LineageEntityV1 = Field(description="Dataset with target columns")
    fields: dict[str, LineageSourceColumnV1] = Field(
        description="Map of target and source columns with type of direct interaction",
        default_factory=dict,
    )


class IndirectLineageColumnRelationV1(BaseModel):
    from_: LineageEntityV1 = Field(description="Dataset with source columns", serialization_alias="from")
    to: LineageEntityV1 = Field(description="Dataset with target columns")
    fields: list[LineageSourceColumnV1] = Field(
        description="List source columns with type of indirect interaction",
        default_factory=list,
    )


class LineageRelationsResponseV1(BaseModel):
    parents: list[LineageParentRelationV1] = Field(description="Parent relations", default_factory=list)
    symlinks: list[LineageSymlinkRelationV1] = Field(description="Symlink relations", default_factory=list)
    inputs: list[LineageInputRelationV1] = Field(description="Input relations", default_factory=list)
    outputs: list[LineageOutputRelationV1] = Field(description="Input relations", default_factory=list)
    direct_column_lineage: list[DirectLineageColumnRelationV1] = Field(
        description="Direct column lineage relations",
        default_factory=list,
    )
    indirect_column_lineage: list[IndirectLineageColumnRelationV1] = Field(
        description="Indirect column lineage relations",
        default_factory=list,
    )


class LineageNodesResponseV1(BaseModel):
    datasets: dict[str, DatasetResponseV1] = Field(description="Dataset nodes", default_factory=dict)
    jobs: dict[str, JobResponseV1] = Field(description="Job nodes", default_factory=dict)
    runs: dict[UUID, RunResponseV1] = Field(description="Run nodes", default_factory=dict)
    operations: dict[UUID, OperationResponseV1] = Field(description="Operation nodes", default_factory=dict)


class LineageResponseV1(BaseModel):
    relations: LineageRelationsResponseV1 = Field(
        description="Lineage relations",
        default_factory=LineageRelationsResponseV1,
    )
    nodes: LineageNodesResponseV1 = Field(description="Lineage nodes", default_factory=LineageNodesResponseV1)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""OpenLineage models & facets.

Currently, openlineage-python [does not support](https://github.com/OpenLineage/OpenLineage/issues/2629) deserialization from JSON.
So we have to write our own deserialization logic.

Also FastStream support only ``pydantic`` models whether openlineage-python provides ``attrs`` models.
"""  # noqa: E501



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create operation

Revision ID: 0b9aac68402b
Revises: 5f8fff06dd76
Create Date: 2024-06-27 19:15:50.909604

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0b9aac68402b"
down_revision = "5f8fff06dd76"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "operation",
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("id", sa.UUID(), nullable=False),
        sa.Column("run_id", sa.UUID(), nullable=False),
        sa.Column("status", sa.SmallInteger(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("type", sa.String(length=32), nullable=False),
        sa.Column("position", sa.Integer(), nullable=True),
        sa.Column("group", sa.String(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("started_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("ended_at", sa.DateTime(timezone=True), nullable=True),
        sa.PrimaryKeyConstraint("created_at", "id", name=op.f("pk__operation")),
        postgresql_partition_by="RANGE (created_at)",
    )
    op.create_index(op.f("ix__operation__run_id"), "operation", ["run_id"], unique=False)


def downgrade() -> None:
    op.drop_index(op.f("ix__operation__run_id"), table_name="operation")
    op.drop_table("operation")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from urllib.parse import quote

from packaging.version import Version

from data_rentgen.consumer.extractors.job import extract_job, extract_parent_job
from data_rentgen.consumer.openlineage.run_event import (
    OpenLineageRunEvent,
    OpenLineageRunEventType,
)
from data_rentgen.consumer.openlineage.run_facets import (
    OpenLineageAirflowDagRunFacet,
    OpenLineageAirflowDagRunType,
    OpenLineageAirflowTaskRunFacet,
    OpenLineageParentRunFacet,
)
from data_rentgen.dto import RunDTO, RunStartReasonDTO, RunStatusDTO
from data_rentgen.dto.user import UserDTO


def extract_parent_run(
    facet: OpenLineageParentRunFacet,
) -> RunDTO:
    return RunDTO(
        id=facet.run.runId,  # type: ignore [arg-type]
        job=extract_parent_job(facet.job),
    )


def extract_run(event: OpenLineageRunEvent) -> RunDTO:
    run = RunDTO(
        id=event.run.runId,  # type: ignore [arg-type]
        job=extract_job(event.job),
    )
    enrich_run_parent(run, event)
    enrich_run_status(run, event)
    enrich_run_start_reason(run, event)
    enrich_run_identifiers(run, event)
    enrich_run_logs(run, event)
    enrich_run_user(run, event)
    return run


def enrich_run_parent(run: RunDTO, event: OpenLineageRunEvent) -> RunDTO:
    if event.run.facets.parent:
        run.parent_run = extract_parent_run(event.run.facets.parent)
    return run


def enrich_run_status(run: RunDTO, event: OpenLineageRunEvent) -> RunDTO:
    match event.eventType:
        case OpenLineageRunEventType.START:
            run.started_at = event.eventTime
            run.status = RunStatusDTO.STARTED
        case OpenLineageRunEventType.RUNNING:
            run.status = RunStatusDTO.STARTED
        case OpenLineageRunEventType.COMPLETE:
            run.ended_at = event.eventTime
            run.status = RunStatusDTO.SUCCEEDED
        case OpenLineageRunEventType.FAIL:
            run.ended_at = event.eventTime
            run.status = RunStatusDTO.FAILED
        case OpenLineageRunEventType.ABORT:
            run.ended_at = event.eventTime
            run.status = RunStatusDTO.KILLED
        case OpenLineageRunEventType.OTHER:
            # OTHER is used only to update run statistics
            pass
    return run


def enrich_run_identifiers(run: RunDTO, event: OpenLineageRunEvent) -> RunDTO:
    spark_application_details = event.run.facets.spark_applicationDetails
    if spark_application_details:
        run.external_id = spark_application_details.applicationId

    airflow_dag_run_facet = event.run.facets.airflowDagRun
    if airflow_dag_run_facet:
        run.external_id = airflow_dag_run_facet.dagRun.run_id

    airflow_task_run_facet = event.run.facets.airflow
    if airflow_task_run_facet:
        run.external_id = airflow_task_run_facet.dagRun.run_id
        run.attempt = str(airflow_task_run_facet.taskInstance.try_number)
    return run


def enrich_run_logs(run: RunDTO, event: OpenLineageRunEvent) -> RunDTO:  # noqa: C901, PLR0912
    spark_application_details = event.run.facets.spark_applicationDetails
    if spark_application_details:
        if spark_application_details.proxyUrl:
            run.running_log_url = spark_application_details.proxyUrl
        else:
            run.running_log_url = spark_application_details.uiWebUrl
        run.persistent_log_url = spark_application_details.historyUrl

        if run.running_log_url:
            run.running_log_url = run.running_log_url.split(",")[0]
        if run.persistent_log_url:
            run.persistent_log_url = run.persistent_log_url.split(",")[0]
        return run

    airflow_dag_run_facet = event.run.facets.airflowDagRun
    if airflow_dag_run_facet:
        namespace = event.job.namespace
        if not namespace.startswith("http"):
            return run

        processing_engine = event.run.facets.processing_engine
        if processing_engine and processing_engine.version >= Version("2.3.0"):
            run.persistent_log_url = get_airflow_2_3_plus_dag_run_url(
                namespace,
                airflow_dag_run_facet,
            )
        else:
            run.persistent_log_url = get_airflow_2_x_dag_run_url(
                namespace,
                airflow_dag_run_facet,
            )

    airflow_task_run_facet = event.run.facets.airflow
    if airflow_task_run_facet:
        # https://github.com/OpenLineage/OpenLineage/pull/2852
        if airflow_task_run_facet.taskInstance.log_url:
            run.persistent_log_url = airflow_task_run_facet.taskInstance.log_url
            return run

        namespace = event.job.namespace
        if not namespace.startswith("http"):
            return run

        processing_engine = event.run.facets.processing_engine
        if processing_engine and processing_engine.version >= Version("2.9.1"):
            run.persistent_log_url = get_airflow_2_9_plus_task_log_url(
                namespace,
                airflow_task_run_facet,
            )
        else:
            run.persistent_log_url = get_airflow_2_x_task_log_url(
                namespace,
                airflow_task_run_facet,
            )
    return run


def get_airflow_2_3_plus_dag_run_url(
    namespace: str,
    airflow_dag_run_facet: OpenLineageAirflowDagRunFacet,
) -> str:
    # https://github.com/apache/airflow/pull/20730
    # https://github.com/apache/airflow/blob/2.9.2/airflow/www/views.py#L2788
    dag_id = airflow_dag_run_facet.dag.dag_id
    dag_run_id = quote(airflow_dag_run_facet.dagRun.run_id)
    return f"{namespace}/dags/{dag_id}/grid?dag_run_id={dag_run_id}"


def get_airflow_2_x_dag_run_url(
    namespace: str,
    airflow_dag_run_facet: OpenLineageAirflowDagRunFacet,
) -> str:
    # https://github.com/apache/airflow/blob/2.9.2/airflow/www/views.py#L2975
    dag_id = quote(airflow_dag_run_facet.dag.dag_id)
    execution_date = quote(airflow_dag_run_facet.dagRun.data_interval_start.isoformat())
    return f"{namespace}/graph?dag_id={dag_id}&execution_date={execution_date}"


def get_airflow_2_9_plus_task_log_url(
    namespace: str,
    airflow_task_run_facet: OpenLineageAirflowTaskRunFacet,
) -> str:
    # https://github.com/apache/airflow/pull/39183
    # https://github.com/apache/airflow/blob/2.9.1/airflow/models/taskinstance.py#L1720-L1734
    dag_id = airflow_task_run_facet.dag.dag_id
    dag_run_id = quote(airflow_task_run_facet.dagRun.run_id)
    task_id = quote(airflow_task_run_facet.task.task_id)
    map_index = airflow_task_run_facet.taskInstance.map_index
    if map_index is None:
        map_index = -1
    return f"{namespace}/dags/{dag_id}/grid?tab=logs&dag_run_id={dag_run_id}&task_id={task_id}&map_index={map_index}"


def get_airflow_2_x_task_log_url(
    namespace: str,
    airflow_task_run_facet: OpenLineageAirflowTaskRunFacet,
) -> str:
    # https://github.com/apache/airflow/blob/2.1.0/airflow/models/taskinstance.py#L524-L528
    dag_id = quote(airflow_task_run_facet.dag.dag_id)
    execution_date = quote(
        airflow_task_run_facet.dagRun.data_interval_start.isoformat(),
    )
    task_id = quote(airflow_task_run_facet.task.task_id)
    return f"{namespace}/log?&dag_id={dag_id}&task_id={task_id}&execution_date={execution_date}"


def enrich_run_start_reason(run: RunDTO, event: OpenLineageRunEvent) -> RunDTO:
    airflow_run_facet = event.run.facets.airflow or event.run.facets.airflowDagRun
    if airflow_run_facet:
        if airflow_run_facet.dagRun.run_type == OpenLineageAirflowDagRunType.MANUAL:
            run.start_reason = RunStartReasonDTO.MANUAL
        else:
            run.start_reason = RunStartReasonDTO.AUTOMATIC
    # For Spark session we cannot determine start reason
    return run


def enrich_run_user(run: RunDTO, event: OpenLineageRunEvent) -> RunDTO:
    spark_application_details = event.run.facets.spark_applicationDetails
    if spark_application_details:
        run.user = UserDTO(name=spark_application_details.userName)

    # Airflow DAG and task have 'owner' field, but if can be either user or group name,
    # and also it does not mean that this exact user started this run.
    # Airflow using different facets for version above provider-opelineage/1.11.0.
    airflow_application_details = event.run.facets.airflow
    if airflow_application_details and all(
        (
            airflow_application_details.dag.owner is not None,
            airflow_application_details.dag.owner != "airflow",
        ),
    ):
        run.user = UserDTO(name=airflow_application_details.dag.owner)  # type: ignore[arg-type]

    airflow_application_dag_details = event.run.facets.airflowDagRun
    if airflow_application_dag_details and all(
        (
            airflow_application_dag_details.dag.owner is not None,
            airflow_application_dag_details.dag.owner != "airflow",
        ),
    ):
        run.user = UserDTO(name=airflow_application_dag_details.dag.owner)  # type: ignore[arg-type]

    return run



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

import textwrap
from enum import Enum

from pydantic import BaseModel, Field

from data_rentgen.consumer.settings.security import KafkaSecurityAnonymousSettings, KafkaSecuritySettings


class KafkaCompression(str, Enum):
    GZIP = "gzip"
    SNAPPY = "snappy"
    LZ4 = "lz4"
    ZSTD = "zstd"

    def __str__(self):
        return self.value


class KafkaSettings(BaseModel):
    """Data.Rentgen consumer Kafka-specific settings.

    These options are passed directly to
    `AIOKafkaConsumer <https://aiokafka.readthedocs.io/en/stable/api.html#aiokafka.AIOKafkaConsumer>`_.

    Examples
    --------

    .. code-block:: bash

        DATA_RENTGEN__KAFKA__BOOTSTRAP_SERVERS=["localhost:9092"]
        DATA_RENTGEN__KAFKA__SECURITY__TYPE=SCRAM-SHA-256
        DATA_RENTGEN__KAFKA__REQUEST_TIMEOUT_MS=5000
        DATA_RENTGEN__KAFKA__CONNECTIONS_MAX_IDLE_MS=540000
    """

    bootstrap_servers: list[str] = Field(
        description="List of Kafka bootstrap servers.",
        min_length=1,
    )
    security: KafkaSecuritySettings = Field(
        default_factory=KafkaSecurityAnonymousSettings,
        description="Kafka security settings.",
    )
    compression: KafkaCompression | None = Field(
        default=None,
        description="Kafka message compression type.",
    )
    # Defaults are copied from FastStream: https://github.com/airtai/faststream/blob/0.5.33/faststream/kafka/fastapi/fastapi.py#L78
    # But only options, related to consuming messages
    request_timeout_ms: int = Field(
        default=40 * 1000,
        description="Client request timeout in milliseconds.",
    )
    retry_backoff_ms: int = Field(
        default=100,
        description="Milliseconds to backoff when retrying on errors.",
    )
    metadata_max_age_ms: int = Field(
        default=5 * 60 * 1000,
        description=textwrap.dedent(
            """
            The period of time in milliseconds after which we force a refresh of metadata,
            even if we haven't seen any partition leadership changes,
            to proactively discover any new brokers or partitions.
            """,
        ),
    )
    connections_max_idle_ms: int = Field(
        default=9 * 60 * 1000,
        description=textwrap.dedent(
            """
            Close idle connections after the number of milliseconds specified by this config.
            Specifying ``None`` will disable idle checks.
            """,
        ),
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.run_facets.base import OpenLineageRunFacet
from data_rentgen.utils import UUIDv6Plus


class OpenLineageParentJob(OpenLineageBase):
    """Parent job identifier.
    See [ParentRunFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/ParentRunFacet.json).
    """

    namespace: str
    name: str


class OpenLineageParentRun(OpenLineageBase):
    """Parent run identifier.
    See [ParentRunFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/ParentRunFacet.json).
    """

    runId: UUIDv6Plus


class OpenLineageParentRunFacet(OpenLineageRunFacet):
    """Run facet describing parent run.
    See [ParentRunFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/ParentRunFacet.json).
    """

    job: OpenLineageParentJob
    run: OpenLineageParentRun



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.openlineage.dataset_facets.base import (
    OpenLineageDatasetFacet,
)


class OpenLineageDatasourceDatasetFacet(OpenLineageDatasetFacet):
    """Dataset facet describing data source information.
    See [DatasourceDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/DatasourceDatasetFacet.json).
    """

    name: str
    uri: str



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from enum import Flag
from uuid import UUID

from sqlalchemy import UUID as SQL_UUID
from sqlalchemy import (
    BigInteger,
    Column,
    Index,
    SmallInteger,
    String,
    func,
)
from sqlalchemy.orm import Mapped, mapped_column

from data_rentgen.db.models.base import Base


class DatasetColumnRelationType(Flag):
    # See https://github.com/OpenLineage/OpenLineage/blob/1.27.0/integration/spark/shared/src/main/java/io/openlineage/spark/agent/lifecycle/plan/column/TransformationInfo.java#L30-L40
    # Using IntFlag to avoid messing up with ARRAY type, bitwise OR is enough
    UNKNOWN = 1

    # Direct
    IDENTITY = 2
    TRANSFORMATION = 4
    TRANSFORMATION_MASKING = 8
    AGGREGATION = 16
    AGGREGATION_MASKING = 32

    # Indirect
    FILTER = 64
    JOIN = 128
    GROUP_BY = 256
    SORT = 512
    WINDOW = 1024
    CONDITIONAL = 2048


# no foreign keys to avoid scanning all the partitions
class DatasetColumnRelation(Base):
    __tablename__ = "dataset_column_relation"
    __table_args__ = (
        Index(
            None,
            Column("fingerprint"),
            Column("source_column"),
            # NULLs are distinct by default, we have to convert them to something else.
            # This is mostly for compatibility with PG <15, there is no `NULLS NOT DISTINCT` option
            func.coalesce(Column("target_column"), ""),
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    fingerprint: Mapped[UUID] = mapped_column(
        SQL_UUID,
        index=False,
        nullable=False,
        doc="Schema SHA-1 digest based used for grouping relations together. Currently this is in form of UUID",
    )

    source_column: Mapped[str] = mapped_column(
        String(length=255),
        index=False,
        nullable=False,
        doc="Source dataset column the data is originated from",
    )

    target_column: Mapped[str | None] = mapped_column(
        String(length=255),
        index=False,
        nullable=True,
        doc=(
            "Target dataset column the data is saved to. NULL means the entire target dataset depends on source column"
        ),
    )

    type: Mapped[DatasetColumnRelationType] = mapped_column(
        SmallInteger(),
        index=False,
        nullable=False,
        doc="Column transformation type",
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.dataset_facets.base import (
    OpenLineageDatasetFacet,
)


class OpenLineageSchemaField(OpenLineageBase):
    """Dataset field information.
    See [SchemaDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/SchemaDatasetFacet.json).
    """

    name: str
    type: str | None = None
    description: str | None = None
    fields: list[OpenLineageSchemaField] | None = None


class OpenLineageSchemaDatasetFacet(OpenLineageDatasetFacet):
    """Dataset facet describing schema.
    See [SchemaDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/SchemaDatasetFacet.json).
    """

    fields: list[OpenLineageSchemaField]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from sqlalchemy import select

from data_rentgen.db.models import User
from data_rentgen.db.repositories.base import Repository
from data_rentgen.dto import UserDTO


class UserRepository(Repository[User]):
    async def get_or_create(self, user: UserDTO) -> User:
        result = await self._get(user.name)
        if not result:
            await self._lock(user.name)
            result = await self._get(user.name) or await self._create(user)
        return result

    async def read_by_id(self, id_: int) -> User | None:
        statement = select(User).where(User.id == id_)
        return await self._session.scalar(statement)

    async def _get(self, name: str) -> User | None:
        statement = select(User).where(User.name == name)
        return await self._session.scalar(statement)

    async def _create(self, user: UserDTO) -> User:
        result = User(name=user.name)
        self._session.add(result)
        await self._session.flush([result])
        return result



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from enum import Enum

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.dataset_facets.base import (
    OpenLineageDatasetFacet,
)


class OpenLineageDatasetLifecycleStateChange(str, Enum):
    """Lifecycle state change type.
    See [LifecycleStateChangeDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/LifecycleStateChangeDatasetFacet.json).
    """

    ALTER = "ALTER"
    CREATE = "CREATE"
    DROP = "DROP"
    OVERWRITE = "OVERWRITE"
    RENAME = "RENAME"
    TRUNCATE = "TRUNCATE"

    def __str__(self) -> str:
        return self.value


class OpenLineageDatasetPreviousIdentifier(OpenLineageBase):
    """Previous identifier information. Used only if `lifecycleStateChange=RENAME`.
    See [LifecycleStateChangeDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/LifecycleStateChangeDatasetFacet.json).
    """

    namespace: str
    name: str


class OpenLineageLifecycleStateChangeDatasetFacet(OpenLineageDatasetFacet):
    """Dataset facet describing lifecycle state change.
    See [LifecycleStateChangeDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/LifecycleStateChangeDatasetFacet.json).
    """

    lifecycleStateChange: OpenLineageDatasetLifecycleStateChange
    previousIdentifier: OpenLineageDatasetPreviousIdentifier | None = None



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Collection

from sqlalchemy import (
    ColumnElement,
    CompoundSelect,
    Select,
    SQLColumnExpression,
    any_,
    asc,
    desc,
    func,
    select,
    union,
)
from sqlalchemy.orm import selectinload

from data_rentgen.db.models import Address, Location
from data_rentgen.db.repositories.base import Repository
from data_rentgen.db.utils.search import make_tsquery, ts_match, ts_rank
from data_rentgen.dto import LocationDTO, PaginationDTO
from data_rentgen.exceptions.entity import EntityNotFoundError


class LocationRepository(Repository[Location]):
    async def create_or_update(self, location: LocationDTO) -> Location:
        result = await self._get(location)
        if not result:
            # try one more time, but with lock acquired.
            # if another worker already created the same row, just use it. if not - create with holding the lock.
            await self._lock(location.type, location.name)
            result = await self._get(location)

        if not result:
            return await self._create(location)

        await self._update_addresses(result, location)
        return result

    async def paginate(
        self,
        page: int,
        page_size: int,
        location_ids: Collection[int],
        location_type: str | None,
        search_query: str | None,
    ) -> PaginationDTO[Location]:
        where = []

        if location_ids:
            where.append(Location.id == any_(list(location_ids)))  # type: ignore[arg-type]

        if location_type:
            where.append(Location.type == location_type)

        query: Select | CompoundSelect
        order_by: list[ColumnElement | SQLColumnExpression]
        if search_query:
            tsquery = make_tsquery(search_query)

            location_stmt = select(Location, ts_rank(Location.search_vector, tsquery).label("search_rank")).where(
                ts_match(Location.search_vector, tsquery),
                *where,
            )
            address_stmt = (
                select(Location, func.max(ts_rank(Address.search_vector, tsquery)).label("search_rank"))
                .join(Location, Address.location_id == Location.id)
                .where(ts_match(Address.search_vector, tsquery), *where)
                .group_by(Location.id, Address.id)
            )

            union_cte = union(location_stmt, address_stmt).cte()

            location_columns = [column for column in union_cte.columns if column.name != "search_rank"]

            query = select(
                *location_columns,
                func.max(union_cte.c.search_rank).label("search_rank"),
            ).group_by(*location_columns)
            order_by = [desc("search_rank"), asc("name"), asc("type")]
        else:
            query = select(Location).where(*where)
            order_by = [Location.name, Location.type]

        options = [selectinload(Location.addresses)]
        return await self._paginate_by_query(
            query=query,
            order_by=order_by,
            options=options,
            page=page,
            page_size=page_size,
        )

    async def update_external_id(self, location_id: int, external_id: str | None) -> Location:
        query = select(Location).where(Location.id == location_id).options(selectinload(Location.addresses))
        location = await self._session.scalar(query)
        if not location:
            raise EntityNotFoundError("Location", "id", location_id)  # noqa: EM101
        location.external_id = external_id
        await self._session.flush([location])
        return location

    async def _get(self, location: LocationDTO) -> Location | None:
        by_name = select(Location).where(Location.type == location.type, Location.name == location.name)
        by_addresses = (
            select(Location)
            .join(Location.addresses)
            .where(
                Location.type == location.type,
                Address.url == any_(list(location.addresses)),  # type: ignore[arg-type]
            )
        )
        statement = (
            select(Location).from_statement(by_name.union(by_addresses)).options(selectinload(Location.addresses))
        )

        return await self._session.scalar(statement)

    async def _create(self, location: LocationDTO) -> Location:
        result = Location(type=location.type, name=location.name)
        self._session.add(result)
        await self._session.flush([result])
        return result

    async def _update_addresses(self, existing: Location, new: LocationDTO) -> Location:
        existing_urls = {address.url for address in existing.addresses}
        new_urls = new.addresses - existing_urls
        # in most cases, Location is unchanged, so we can avoid UPDATE statements
        if not new_urls:
            return existing

        # take a lock, to avoid race conditions, and then
        # get fresh state of the object, because it already could be updated by another worker
        await self._lock(existing.type, existing.name)
        await self._session.refresh(existing, ["addresses"])

        # already has all required addresses - nothing to update
        existing_urls = {address.url for address in existing.addresses}
        new_urls = new.addresses - existing_urls
        if not new_urls:
            return existing

        # add new addresses while holding the lock
        addresses = [Address(url=url, location_id=existing.id) for url in new_urls]
        existing.addresses.extend(addresses)
        await self._session.flush([existing])
        return existing
